/* ============================================================================== 
 * Copyright (c) 2013 Pengchong Jin (Purdue University)
 * Copyright (c) 2013 S. Jordan Kisner (Purdue University)
 * Copyright (c) 2013 Thomas Frese
 * Copyright (c) 2013 Charles A. Bouman (Purdue University)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Pengchong Jin, S. J. Kisner, Thomas Frese, Charles A. Bouman
 * Purdue University, nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ============================================================================== */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>	/* strcmp */
#include <math.h>

#include "data2d.h"
#include "allocate.h"
#include "A_comp.h"

#define STORE_A_MATRIX
#define WIDE_BEAM

/******************************************************************/
/* Compute line segment lengths through a pixel for the given set */
/* of sinogram angles and for multiple displacements (LEN_PIX)    */
/******************************************************************/
/* Fill profiles of pixels from all angles
   ONLY FOR SQUARE PIXELS NOW   
   Each profile assumed 2 pixels wide
   This is used to speed calculation (too many to store) of the
   entries of the projection matrix.
   Values are scaled by "scale"
 */
void pix_prof_comp(sino_info, img_info, pix_prof)
	struct SinoInfo *sino_info;
	struct ImgInfo *img_info;
	ENTRY ***pix_prof;
{
	int i, j;
	ENTRY pi, ang, d1, d2, t, t_1, t_2, t_3, t_4, maxval, rc, delta;

	delta = img_info->delta;

	*pix_prof = (ENTRY **)get_img(LEN_PIX, sino_info->N_theta, sizeof(ENTRY));

	pi = 4.0*atan(1.0);

	rc = sin(pi/4.0);
	for (i = 0; i < sino_info->N_theta; i++)
	{
		ang = sino_info->theta_0 + i*sino_info->delta_theta;

		while (ang >= pi/2.0)
		{
			ang -= pi/2.0;
		}
		while(ang < 0.0)
		{
			ang += pi/2.0;
		}

		if (ang <= pi/4.0)
		{
			maxval = delta/cos(ang);
		}
		else
		{
			maxval = delta/cos(pi/2.0-ang);
		}

		d1 = rc*cos(pi/4.0-ang);
		d2 = rc*fabs(sin(pi/4.0-ang));

		t_1 = 1.0 - d1;
		t_2 = 1.0 - d2;
		t_3 = 1.0 + d2;
		t_4 = 1.0 + d1;
		for (j = 0; j < LEN_PIX; j++)
		{
			t = 2.0*j/(ENTRY)LEN_PIX;
			if (t <= t_1 || t > t_4)
			{
				(*pix_prof)[i][j] = 0.0;
			}
			else if (t <= t_2)
			{
				(*pix_prof)[i][j] = maxval*(t-t_1)/(t_2-t_1);
			}
			else if (t <= t_3)
			{
				(*pix_prof)[i][j] = maxval;
			}
			else
			{
				(*pix_prof)[i][j] = maxval*(t_4-t)/(t_4-t_3);
			}
		}
	}
}

void A_comp_ij(
	int im_row,
	int im_col,
	struct SinoInfo *sino_info,
	struct ImgInfo *img_info,
	ENTRY **pix_prof,
	struct ACol *A_col)
{
	int ind, ind_min, ind_max, pr;
	int prof_ind, pix_prof_ind, pind, i, proj_count;
	ENTRY Aval, t_min, t_max, ang, x, y;
	int k;
	ENTRY d1, d2, z, t, const1, const2, const3, const4, sum;
	static int n_theta, n_t, n_x, n_y;
	static ENTRY delta, delta_theta, delta_t, theta_0, t_0, x_0, y_0;

	ENTRY d_detector = 1000;
	ENTRY d_source = 1000;
	static ENTRY dprof[LEN_DET];
	static int first = 1;

	t = 0;
	prof_ind = 0;
	pix_prof_ind = 0;
	sum = 0;

	if (first == 1)
	{
		first = 0;

		n_theta = sino_info->N_theta;
		n_t = sino_info->N_t;
		delta_theta = sino_info->delta_theta;
		delta_t = sino_info->delta_t;
		theta_0 = sino_info->theta_0;
		t_0 = sino_info->t_0;

		n_x = img_info->Nx;
		n_y = img_info->Ny;
		delta = img_info->delta;
		x_0 = img_info->x_0;
		y_0 = img_info->y_0;

		/* delta profile */
		/*
		   for(k=0;k<LEN_DET;k++) 
		   dprof[k] =0;
		   dprof[(LEN_DET-1)/2]=1.0;
		 */

		/* square profile */
		for (k = 0; k < LEN_DET; k++)
		{
			dprof[k] = 1.0/(LEN_DET);
		}

		/* triangular profile */
		/*
		   sum=0;
		   for(k=0;k<LEN_DET;k++) {
		   if(k<=(LEN_DET-1)/2)
		   dprof[k]=2.0*k/(LEN_DET-1);
		   else
		   dprof[k]=2.0*(LEN_DET-1-k)/(LEN_DET-1);
		   sum+=dprof[k];
		   }
		   for(k=0;k<LEN_DET;k++) 
		   dprof[k] /= sum;
		 */
	} 

	/* WATCH THIS; ONLY FOR SQUARE PIXELS NOW   */

	y = y_0 + im_row*delta;
	x = x_0 + im_col*delta;

	proj_count = 0;
	for (pr = 0; pr < n_theta; pr++)
	{
		pind = pr*n_t;
		ang = theta_0 + pr*(delta_theta);

		z = x*cos(ang) + y*sin(ang);
		d1 = d_source + z;
		d2 = d_detector - z;

		/* range for pixel profile.  Need profile to contain 2 pixel widths */
		t_min = y*cos(ang) - x*sin(ang) - delta;
		t_max = t_min + 2.0*delta;
		/* This also prevents over-reach (with rounding of negative numbers)  */
		if (t_max < t_0) 
		{
			continue;
		}

		/* Changed 10/2000 (TF): more general, old implementation assumed
		   that geom->dia=sino->n_t*sino->delta_t; */ 

#ifdef WIDE_BEAM
		/* Relevant detector indices */
		ind_min = ceil((t_min-t_0)/delta_t - 0.5);
		ind_max = (t_max-t_0)/delta_t + 0.5;
#else 
		ind_min = ceil((t_min-t_0)/delta_t);
		ind_max = (t_max-t_0)/delta_t;
#endif

		/* Fix this 4/91 to prevent over-reach at ends  */
		ind_min = (ind_min<0) ? 0 : ind_min;
		ind_max = (ind_max>=n_t) ? n_t-1 : ind_max;

		const1 = t_0 - delta_t/2.0*d1/(d1+d2);
		const2 = delta_t/(ENTRY)(LEN_DET-1)*d1/(d1+d2);
		const3 = delta - (y*cos(ang) - x*sin(ang));
		const4 = (ENTRY)(LEN_PIX-1)/(2.0*delta);

		for (i = ind_min; i <= ind_max; i++)
		{
			ind = pind + i;

#ifdef WIDE_BEAM
			/* step through values of detector profile, inner product with PIX prof */
			Aval = 0;
			for (k = 0; k < LEN_DET; k++)
			{
				t = const1 + (ENTRY)i*delta_t + (ENTRY)k*const2;
				pix_prof_ind = (t+const3)*const4 +0.5;   /* +0.5 for rounding */
				if (pix_prof_ind >= 0 && pix_prof_ind < LEN_PIX)
				{
					Aval+= dprof[k]*pix_prof[pr][pix_prof_ind];
				}
			}
#else
			/*** this block computes zero-beam-width projection model ****/
			prof_ind = LEN_PIX*(t_0+i*delta_t-t_min)/(2.0*delta);

			if (prof_ind >= LEN_PIX || prof_ind < 0)
			{
				if (prof_ind == LEN_PIX)
				{
					prof_ind = LEN_PIX-1;
				}
				else if (prof_ind == -1)
				{
					prof_ind = 0;
				}
				else
				{
					fprintf(stderr,"\nExiting Program: input parameters inconsistant\n");
					exit(-1);
				}
			}
			Aval = pix_prof[pr][prof_ind];
#endif

			if (Aval > 0.0)
			{
				A_col->val[proj_count] = Aval;
				A_col->index[proj_count] = ind;   /* Index for ray */
				proj_count++;
			}
		}
	} 

	A_col->n_index = proj_count;
}

void A_comp(
	int im_row,
	int im_col,
	struct SinoInfo *sino_info,
	struct ImgInfo *img_info,
	ENTRY **pix_prof,
	struct ACol *A_col)
{

#ifdef STORE_A_MATRIX
/* precompute entire A-matrix at first call, then 
   just return precomputed cols in subsequent calls */

	static int first = 1;
	int i, j, r, nr;
	int col_length, n_x, n_y;

	static struct ACol A_col_sgl;
	static struct ACol **A_col_arr;
	static struct ACol *A_col_ptr;

	if (first == 1)
	{
		first = 0;

		fprintf(stdout, "Initializing A-matrix...\n");
		fflush(stdout);

		col_length = sino_info->N_t*sino_info->N_theta;

		A_col_sgl.val = (ENTRY *)get_spc(col_length, sizeof(ENTRY));
		A_col_sgl.index = (int *)get_spc(col_length, sizeof(int));

		n_x = img_info->Nx;
		n_y = img_info->Ny;
		A_col_arr = (struct ACol **)multialloc(sizeof(struct ACol), 2, n_y, n_x);

		for (i = 0; i < n_y; i++)
		{
			for (j = 0; j < n_x; j++)
			{
				A_comp_ij(i, j, sino_info, img_info, pix_prof, &A_col_sgl);

				A_col_arr[i][j].val = (ENTRY *)get_spc(A_col_sgl.n_index, sizeof(ENTRY));
				A_col_arr[i][j].index = (int *)get_spc(A_col_sgl.n_index, sizeof(int));
				A_col_arr[i][j].n_index = A_col_sgl.n_index;
				for (r = 0; r < A_col_sgl.n_index; r++)
				{
					A_col_arr[i][j].val[r] = (ENTRY)A_col_sgl.val[r];
					A_col_arr[i][j].index[r] = A_col_sgl.index[r];
				}
			}
		}

		free((void *)A_col_sgl.val);
		free((void *)A_col_sgl.index);

		fprintf(stdout, "done\n");
		fflush(stdout);
	}

	A_col_ptr = &A_col_arr[im_row][im_col];
	A_col->n_index = nr = A_col_ptr->n_index;

	for (r = 0; r < nr; r++)
	{
		A_col->val[r] = (ENTRY)A_col_ptr->val[r];
		A_col->index[r] = A_col_ptr->index[r];
	}
#else
	/* compute A_*j on the fly */
	A_comp_ij(im_row, im_col, sino_info, img_info, pix_prof, A_col);
#endif
}


void readAmatrix(
	char *fname,
	struct ACol **A_col_arr,	/* A_col_arr should be allocated first before enter the function */
	struct ImgInfo *img_info)
{
	FILE *fp;
	int jx, jy, Nx, Ny;
	int M_nonzero;

	Nx = img_info->Nx;
	Ny = img_info->Ny;

	if ((fp = fopen(fname, "r")) == NULL)
	{
		fprintf(stderr, "ERROR in readAmatrix: can't open file %s.\n", fname);
		exit(-1);
	}

	for (jy = 0; jy < Ny; jy++)
	{
		for (jx = 0; jx < Nx; jx++)
		{
			fread(&M_nonzero, sizeof(int), 1, fp);

			A_col_arr[jy][jx].val = (ENTRY *)get_spc(M_nonzero, sizeof(ENTRY));
			A_col_arr[jy][jx].index = (int *)get_spc(M_nonzero, sizeof(int));
			A_col_arr[jy][jx].n_index = M_nonzero;

			fread(A_col_arr[jy][jx].index, sizeof(int), M_nonzero, fp);
			fread(A_col_arr[jy][jx].val, sizeof(ENTRY), M_nonzero, fp);
		}
	}

	fclose(fp);
}

void writeAmatrix(
	char *fname,
	struct ACol **A_col_arr,
	struct ImgInfo *img_info)
{
	FILE *fp;
	int jx, jy, Nx, Ny;
	int M_nonzero;

	Nx = img_info->Nx;
	Ny = img_info->Ny;

	if ((fp = fopen(fname, "w")) == NULL)
	{
		fprintf(stderr, "ERROR in writeAmatrix: can't open file %s.\n", fname);
		exit(-1);
	}

	for (jy = 0; jy < Ny; jy++)
	{
		for (jx = 0; jx < Nx; jx++)
		{
			M_nonzero = A_col_arr[jy][jx].n_index;
			fwrite(&M_nonzero, sizeof(int), 1, fp);

			fwrite(A_col_arr[jy][jx].index, sizeof(int), M_nonzero, fp);
			fwrite(A_col_arr[jy][jx].val, sizeof(ENTRY), M_nonzero, fp);
		}
	}

	fclose(fp);
}
