/* ============================================================================== 
 * Copyright (c) 2013 Pengchong Jin (Purdue University)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Pengchong Jin, Purdue University,
 * nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ============================================================================== */

/* ========================================================== */
/*          Definition of Data Structure for MBIR             */
/*           by Pengchong Jin, Purdue University              */
/* ========================================================== */

#ifndef _DATA2D_H_
#define _DATA2D_H_

typedef double	ENTRY;

#define PI	3.1415926535897932384
#define ATT_AIR	0.0
#define ATT_H2O	2.02527e-2
#define hu2mu(hu, att_air, att_h2o) (hu*(att_h2o-att_air)/1000.0+att_air)
#define mu2hu(mu, att_air, att_h2o) (1000.0*(mu-att_air)/(att_h2o-att_air))

struct ImgInfo
{
	int Nx;
	int Ny;
	ENTRY fov;
	ENTRY r;
	ENTRY delta;
	ENTRY x_0;
	ENTRY y_0;
	ENTRY mu_water;		/* mm-1 */
	int upsample;
	char imgFile[200];
};

struct Image
{
	struct ImgInfo img_info;
	char **recon_mask;
	ENTRY *img;
};

struct SinoInfo
{
	int N_theta;
	int N_t;
	ENTRY delta_theta;
	ENTRY delta_t;
	ENTRY theta_0;
	ENTRY t_0;
	ENTRY norm_factor;
	char sinoFile[200];
};

struct Sinogram
{
	struct SinoInfo sino_info;
	ENTRY *sino;
};

struct ACol
{
	int n_index;
	int *index;
	ENTRY *val;
};

struct PriorInfo
{
	char est;	/* 0 for ML, 1 for quad, 2 for q-GGMRF */
	ENTRY p;
	ENTRY q;
	ENTRY c;
	ENTRY sigma;
	int weight;	/* 0 for identity, 1 for exponential */
	ENTRY stop_threshold;
	int Nit;
	char oname[200];
};

#endif
