/* ============================================================================== 
 * Copyright (c) 2013 Pengchong Jin (Purdue University)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Pengchong Jin, Purdue University,
 * nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ============================================================================== */

#include "data2d.h"
#include "A_comp.h"
#include "fpro2d.h"

void forwardProject2D(
	ENTRY *Ax,
	ENTRY *x,
	ENTRY **profiles,
	struct ACol *A_col,
	struct SinoInfo *sino_info,
	struct ImgInfo *img_info)
{
	int jx, jy, Nx, Ny, i, M, r;

	Nx = img_info->Nx;
	Ny = img_info->Ny;
	M = sino_info->N_theta*sino_info->N_t;

	for (i = 0; i < M; i++)
	{
		Ax[i] = 0.0;
	}
	for (jy = 0; jy < Ny; jy++)
	{
		for (jx = 0; jx < Nx; jx++)
		{
			A_comp(jy, jx, sino_info, img_info, profiles, A_col);

			if (A_col->n_index > 0)
			{
				for (r = 0; r < A_col->n_index; r++)
				{
					Ax[A_col->index[r]] += A_col->val[r]*x[jy*Nx+jx];
				}
			}
		}
	}
}
